from typing import Sequence, List, Union, Optional, Tuple

import numpy as np
import pandas as pd
import tensorflow as tf

from sknlp.vocab import Vocab
from .nlp_dataset import NLPDataset
from .utils import serialize_example


class ClassificationDataset(NLPDataset):
    def __init__(
        self,
        vocab: Vocab,
        labels: Sequence[str],
        df: Optional[pd.DataFrame] = None,
        csv_file: Optional[str] = None,
        in_memory: bool = True,
        no_label: bool = False,
        is_multilabel: bool = True,
        max_length: Optional[int] = None,
        text_segmenter: str = "char",
        text_dtype: tf.DType = tf.int32,
        label_dtype: tf.DType = tf.float32,
        batch_padding_shapes: Optional[Tuple[tf.DType]] = ((None,), (None,)),
        batch_padding_values: Optional[Tuple[tf.DType]] = (0, 0.0),
    ):
        self.vocab = vocab
        self.labels = list(labels)
        self.is_multilabel = is_multilabel
        self.label2idx = dict(zip(labels, range(len(labels))))
        super().__init__(
            df=df,
            csv_file=csv_file,
            in_memory=in_memory,
            no_label=no_label,
            text_segmenter=text_segmenter,
            max_length=max_length,
            na_value="" if is_multilabel else "NULL",
            column_dtypes=["str", "str"],
            text_dtype=text_dtype,
            label_dtype=label_dtype,
            batch_padding_shapes=batch_padding_shapes,
            batch_padding_values=batch_padding_values,
        )

    @property
    def y(self) -> Union[List[str], List[List[str]]]:
        if self.no_label:
            return []
        return [
            data[-1].decode("utf-8").split("|")
            if self.is_multilabel
            else data[-1].decode("Utf-8")
            for data in self._original_dataset.as_numpy_iterator()
        ]

    def _text_transform(self, text: tf.Tensor) -> np.ndarray:
        tokens = super()._text_transform(text)
        return np.array(
            [self.vocab[t] for t in tokens[: self.max_length]], dtype=np.int32
        )

    def _label_binarizer(self, labels: List[str]) -> np.ndarray:
        label2idx = self.label2idx
        res = np.zeros(len(label2idx), dtype=np.float32)
        res[[label2idx[label] for label in labels if label in label2idx]] = 1
        return res

    def _label_transform(self, label: tf.Tensor) -> np.ndarray:
        _label = super()._label_transform(label)
        if self.is_multilabel:
            labels = _label.split("|")
        else:
            labels = [_label]
        return self._label_binarizer(labels)

    def to_tfrecord(self, filename: str) -> None:
        def func(text: np.ndarray, label: np.ndarray):
            return tf.reshape(
                serialize_example(
                    (self._text_transform(text), self._label_transform(label)),
                    ("tensor", "tensor"),
                ),
                (),
            )

        tf_writer = tf.data.experimental.TFRecordWriter(filename)
        tf_writer.write(
            self._dataset.map(
                lambda t, l: tf.py_function(func, inp=[t, l], Tout=tf.string),
                num_parallel_calls=tf.data.experimental.AUTOTUNE,
            )
        )

    @classmethod
    def from_tfrecord(cls, filename: str) -> tf.data.Dataset:
        def func(record: tf.Tensor):
            parsed_record = tf.io.parse_single_example(
                record,
                {
                    "feature0": tf.io.FixedLenFeature([], tf.string, default_value=""),
                    "feature1": tf.io.FixedLenFeature([], tf.string, default_value=""),
                },
            )
            return (
                tf.io.parse_tensor(parsed_record["feature0"], tf.int32),
                tf.io.parse_tensor(parsed_record["feature1"], tf.float32),
            )

        return tf.data.TFRecordDataset(filename).map(func)



import tensorflow as tf


class TextRNNLayer(tf.keras.layers.Layer):

    def __init__(
        self,
        num_rnn_layers=1,
        rnn_hidden_size=256,
        rnn_projection_size=128,
        rnn_recurrent_clip=3,
        rnn_projection_clip=3,
        rnn_input_dropout=0.5,
        rnn_recurrent_dropout=0.5,
        rnn_output_dropout=0.5,
        rnn_last_connection=None,
        rnn_kernel_initializer='glorot_uniform',
        rnn_recurrent_initializer='orthogonal',
        rnn_projection_initializer='glorot_uniform',
        rnn_bias_initializer='zeros',
        rnn_kernel_regularizer=None,
        rnn_recurrent_regularizer=None,
        rnn_projection_regularizer=None,
        rnn_bias_regularizer=None,
        rnn_kernel_constraint=None,
        rnn_recurrent_constraint=None,
        rnn_projection_constraint=None,
        rnn_bias_constraint=None,
    ):
        self.num_rnn_layers = num_rnn_layers
        self.rnn_hidden_size = rnn_hidden_size
        self.rnn_last_connection = rnn_last_connection
        self.rnn_projection_size = rnn_projection_size
        self.rnn_recurrent_clip = rnn_recurrent_clip
        self.rnn_projection_clip = rnn_projection_clip
        self.rnn_input_dropout = rnn_input_dropout
        self.rnn_recurrent_dropout = rnn_recurrent_dropout
        self.rnn_output_dropout = rnn_output_dropout
        self.rnn_kernel_initializer = rnn_kernel_initializer
        self.rnn_recurrent_initializer = rnn_recurrent_initializer
        self.rnn_projection_initializer = rnn_projection_initializer
        self.rnn_bias_initializer = rnn_bias_initializer
        self.rnn_kernel_regularizer = rnn_kernel_regularizer
        self.rnn_recurrent_regularizer = rnn_recurrent_regularizer
        self.rnn_projection_regularizer = rnn_projection_regularizer
        self.rnn_bias_regularizer = rnn_bias_regularizer
        self.rnn_kernel_constraint = rnn_kernel_constraint
        self.rnn_recurrent_constraint = rnn_recurrent_constraint
        self.rnn_projection_constraint = rnn_projection_constraint
        self.rnn_bias_constraint = rnn_bias_constraint



from typing import Optional, Dict

import functools
import tensorflow as tf
from tensorflow.keras.metrics import Accuracy, BinaryAccuracy


class AccuracyWithLogits(Accuracy):
    def __init__(
        self,
        name: str = "accuracy",
        dtype: Optional[tf.DType] = None,
        logits2scores: str = "sigmoid",
    ) -> None:
        super().__init__(name=name, dtype=dtype)
        self.logits2scores = logits2scores
        if logits2scores == "sigmoid":
            self._l2s = tf.math.sigmoid
        elif logits2scores == "softmax":
            self._l2s = functools.partial(tf.math.softmax, axis=-1)

    def update_state(
        self,
        y_true: tf.Tensor,
        y_pred: tf.Tensor,
        sample_weight: Optional[tf.Tensor] = None,
    ) -> None:
        super().update_state(y_true, self._l2s(y_pred), sample_weight=sample_weight)

    def get_config(self) -> Dict[str, str]:
        return {**super().get_config(), "logits2scores": self.logits2scores}


class BinaryAccuracyWithLogits(BinaryAccuracy):
    def update_state(
        self,
        y_true: tf.Tensor,
        y_pred: tf.Tensor,
        sample_weight: Optional[tf.Tensor] = None,
    ) -> None:
        super().update_state(
            y_true, tf.math.sigmoid(y_pred), sample_weight=sample_weight
        )



from tensorflow.keras.callbacks import Callback
import tensorflow.keras.backend as K


class WeightDecayScheduler(Callback):
    def __init__(self, schedule, verbose=0):
        super().__init__()
        self.schedule = schedule
        self.verbose = verbose

    def on_epoch_begin(self, epoch, logs=None):
        if not hasattr(self.model.optimizer, "weight_decay"):
            raise ValueError('Optimizer must have a "weight_decay" attribute.')
        wd = float(K.get_value(self.model.optimizer.weight_decay))
        wd = self.schedule(epoch, wd)
        K.set_value(self.model.optimizer.weight_decay, K.get_value(wd))
        if self.verbose > 0:
            print(
                "\nEpoch %05d: WeightDecayScheduler reducing weight decay "
                "to %s." % (epoch + 1, wd)
            )

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        logs["weight_decay"] = K.get_value(self.model.optimizer.weight_decay)



from typing import Sequence, List, Dict, Any, Optional, Union, Tuple

import numpy as np
import pandas as pd
from tabulate import tabulate
import tensorflow as tf
import tensorflow_addons as tfa

from sknlp.vocab import Vocab
from sknlp.data import SimilarityDataset
from sknlp.metrics import BinaryAccuracyWithLogits

from ..supervised_model import SupervisedNLPModel
from ..classifiers.utils import (
    classification_fscore,
    logits2probabilities,
    probabilities2classes,
)
from ..text2vec import Text2vec


class DeepDiscriminator(SupervisedNLPModel):
    def __init__(
        self,
        classes: Sequence[str] = ("相似度",),
        max_sequence_length: Optional[int] = None,
        sequence_length: Optional[int] = None,
        segmenter: Optional[str] = "jieba",
        embedding_size: int = 100,
        use_batch_normalization: bool = True,
        text2vec: Optional[Text2vec] = None,
        loss: Optional[str] = None,
        loss_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs
    ):
        super().__init__(
            list(classes),
            max_sequence_length=max_sequence_length,
            sequence_length=sequence_length,
            segmenter=segmenter,
            embedding_size=embedding_size,
            text2vec=text2vec,
            task="similarity",
            **kwargs
        )
        self._use_batch_normalization = use_batch_normalization
        self._loss = loss
        self._loss_kwargs = loss_kwargs

    @property
    def use_batch_normalization(self):
        return self._use_batch_normalization

    def get_loss(self, *args, **kwargs):
        if self._loss == "focal":
            return tfa.losses.SigmoidFocalCrossEntropy(
                from_logits=True, **self._loss_kwargs
            )
        return tf.keras.losses.BinaryCrossentropy(from_logits=True)

    def get_metrics(self, *args, **kwargs) -> List[tf.keras.metrics.Metric]:
        return [BinaryAccuracyWithLogits()]

    @classmethod
    def get_monitor(cls) -> str:
        return "val_binary_accuracy"

    def create_dataset_from_df(
        self, df: pd.DataFrame, vocab: Vocab, segmenter: str, labels: Sequence[str]
    ) -> SimilarityDataset:
        return SimilarityDataset(
            vocab,
            list(labels),
            df=df,
            max_length=self.max_sequence_length,
            text_segmenter=segmenter,
        )

    @classmethod
    def dummy_y(self, X: Sequence[str]) -> List[float]:
        return [0 for _ in range(len(X))]

    def predict_proba(
        self,
        X: Sequence[Tuple[str, str]] = None,
        *,
        dataset: SimilarityDataset = None,
        batch_size: int = 128
    ) -> np.ndarray:
        logits = super().predict(X=X, dataset=dataset, batch_size=batch_size)
        return logits2probabilities(logits, True)

    def predict(
        self,
        X: Sequence[Tuple[str, str]] = None,
        *,
        dataset: SimilarityDataset = None,
        batch_size: int = 128
    ) -> List[float]:
        return self.predict_proba(X=X, dataset=dataset, batch_size=batch_size)

    def score(
        self,
        X: Sequence[Tuple[str, str]] = None,
        y: Sequence[float] = None,
        *,
        dataset: SimilarityDataset = None,
        threshold: float = 0.5,
        batch_size: int = 128
    ) -> pd.DataFrame:
        dataset = self.prepare_dataset(X, y, dataset)
        probs = self.predict(dataset=dataset, batch_size=batch_size)
        predictions = probabilities2classes(probs, True, thresholds=threshold)
        return classification_fscore(dataset.y, predictions, classes=[0, 1])

    @classmethod
    def format_score(cls, score_df: pd.DataFrame, format: str = "markdown") -> str:
        return tabulate(score_df, headers="keys", tablefmt="github", showindex=False)

    def get_config(self) -> Dict[str, Any]:
        return {
            **super().get_config(),
            "use_batch_normalization": self.use_batch_normalization,
        }

    @classmethod
    def _filter_config(cls, config):
        config = super()._filter_config(config)
        config.pop("algorithm", None)
        config.pop("task", None)
        return config

    @classmethod
    def get_custom_objects(cls) -> Dict[str, Any]:
        return {
            **super().get_custom_objects(),
            "BinaryAccuracyWithLogits": BinaryAccuracyWithLogits,
        }


from typing import Any, Dict, Sequence, Optional

import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Dropout
import pandas as pd

from sknlp.layers import (
    MLPLayer,
    BertLayer,
    BertPreprocessingLayer,
)
from sknlp.vocab import Vocab
from sknlp.data import BertClassificationDataset
from sknlp.module.text2vec import Bert2vec
from .deep_classifier import DeepClassifier


class BertClassifier(DeepClassifier):
    def __init__(
        self,
        classes: Sequence[str],
        is_multilabel: bool = True,
        segmenter: Optional[str] = None,
        embedding_size: int = 100,
        use_batch_normalization: bool = True,
        max_sequence_length: int = 120,
        num_fc_layers: int = 2,
        fc_hidden_size: int = 256,
        fc_activation: str = "tanh",
        fc_momentum: float = 0.9,
        fc_epsilon: float = 1e-5,
        output_dropout: float = 0.5,
        text2vec: Optional[Bert2vec] = None,
        **kwargs
    ) -> None:
        super().__init__(
            classes,
            is_multilabel=is_multilabel,
            segmenter=segmenter,
            algorithm="bert",
            embedding_size=embedding_size,
            use_batch_normalization=use_batch_normalization,
            max_sequence_length=max_sequence_length,
            text2vec=text2vec,
            **kwargs
        )
        self.num_fc_layers = num_fc_layers
        self.fc_hidden_size = fc_hidden_size
        self.fc_activation = fc_activation
        self.fc_momentum = fc_momentum
        self.fc_epsilon = fc_epsilon
        self.output_dropout = output_dropout
        self.inputs = tf.keras.Input(shape=(), dtype=tf.string, name="text_input")

    def create_dataset_from_df(
        self,
        df: pd.DataFrame,
        vocab: Vocab,
        segmenter: str,
        labels: Sequence[str],
        no_label: bool,
    ) -> BertClassificationDataset:
        return BertClassificationDataset(
            vocab,
            list(labels),
            df=df,
            is_multilabel=self.is_multilabel,
            max_length=self.max_sequence_length,
            no_label=no_label,
        )

    def build_encode_layer(self, inputs: tf.Tensor) -> tf.Tensor:
        preprocessing_layer = BertPreprocessingLayer(self.text2vec.vocab.sorted_tokens)
        token_ids = preprocessing_layer(inputs)
        _, _, cls = self.text2vec(
            [
                token_ids,
                K.zeros_like(token_ids, dtype=tf.int64),
            ]
        )
        if self.output_dropout:
            return Dropout(self.output_dropout, name="embedding_dropout")(cls)
        return cls

    def build_output_layer(self, inputs: tf.Tensor) -> tf.Tensor:
        return MLPLayer(
            self.num_fc_layers,
            hidden_size=self.fc_hidden_size,
            output_size=self.num_classes,
            activation=self.fc_activation,
            batch_normalization=self.use_batch_normalization,
            momentum=self.fc_momentum,
            epsilon=self.fc_epsilon,
            name="mlp",
        )(inputs)

    def get_config(self) -> Dict[str, Any]:
        return {**super().get_config(), "output_dropout": self.output_dropout}

    def get_custom_objects(self) -> Dict[str, Any]:
        return {
            **super().get_custom_objects(),
            "MLPLayer": MLPLayer,
            "BertLayer": BertLayer,
            "BertPreprocessingLayer": BertPreprocessingLayer,
        }



from typing import Sequence, List, Dict, Any, Optional, Union, Tuple

import numpy as np
import pandas as pd
from tabulate import tabulate
import tensorflow as tf
import tensorflow_addons as tfa

from sknlp.vocab import Vocab
from sknlp.data import ClassificationDataset
from sknlp.metrics import PrecisionWithLogits, RecallWithLogits, FBetaScoreWithLogits

from ..supervised_model import SupervisedNLPModel
from ..text2vec import Text2vec

from .utils import logits2probabilities, probabilities2classes, classification_fscore


class DeepClassifier(SupervisedNLPModel):
    def __init__(
        self,
        classes: Sequence[str],
        is_multilabel: bool = True,
        max_sequence_length: Optional[int] = None,
        sequence_length: Optional[int] = None,
        segmenter: Optional[str] = "jieba",
        embedding_size: int = 100,
        use_batch_normalization: bool = True,
        text2vec: Optional[Text2vec] = None,
        loss: Optional[str] = None,
        loss_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs
    ):
        classes = list(classes)
        if not is_multilabel and "NULL" != classes[0]:
            if "NULL" in classes:
                classes.remove("NULL")
            classes.insert(0, "NULL")
        super().__init__(
            classes,
            max_sequence_length=max_sequence_length,
            sequence_length=sequence_length,
            segmenter=segmenter,
            embedding_size=embedding_size,
            text2vec=text2vec,
            task="classification",
            **kwargs
        )
        self._is_multilabel = is_multilabel
        self._use_batch_normalization = use_batch_normalization
        self._loss = loss
        self._loss_kwargs = loss_kwargs

    @property
    def is_multilabel(self):
        return self._is_multilabel

    @property
    def use_batch_normalization(self):
        return self._use_batch_normalization

    def get_loss(self, *args, **kwargs):
        if self.is_multilabel:
            if self._loss == "focal":
                return tfa.losses.SigmoidFocalCrossEntropy(
                    from_logits=True, **self._loss_kwargs
                )
            return tf.keras.losses.BinaryCrossentropy(from_logits=True)
        else:
            return tf.keras.losses.CategoricalCrossentropy(from_logits=True)

    def get_metrics(self, *args, **kwargs) -> List[tf.keras.metrics.Metric]:
        if self.is_multilabel:
            return [
                PrecisionWithLogits(),
                RecallWithLogits(),
                FBetaScoreWithLogits(self.num_classes),
            ]
        else:
            return [
                PrecisionWithLogits(logits2scores="softmax"),
                RecallWithLogits(logits2scores="softmax"),
                FBetaScoreWithLogits(self.num_classes, logits2scores="softmax"),
            ]

    @classmethod
    def get_monitor(cls) -> str:
        return "val_fbeta_score"

    def create_dataset_from_df(
        self,
        df: pd.DataFrame,
        vocab: Vocab,
        segmenter: str,
        labels: Sequence[str],
        no_label: bool,
    ) -> ClassificationDataset:
        return ClassificationDataset(
            vocab,
            list(labels),
            df=df,
            is_multilabel=self.is_multilabel,
            max_length=self.max_sequence_length,
            no_label=no_label,
            text_segmenter=segmenter,
        )

    def predict_proba(
        self,
        X: Sequence[str] = None,
        *,
        dataset: ClassificationDataset = None,
        batch_size: int = 128
    ) -> np.ndarray:
        logits = super().predict(X=X, dataset=dataset, batch_size=batch_size)
        return logits2probabilities(logits, self.is_multilabel)

    def predict(
        self,
        X: Sequence[str] = None,
        *,
        dataset: ClassificationDataset = None,
        thresholds: Union[float, List[float]] = 0.5,
        batch_size: int = 128
    ) -> Union[List[str], List[List[str]]]:
        probabilities = self.predict_proba(X=X, dataset=dataset, batch_size=batch_size)
        predictions = probabilities2classes(
            probabilities, self.is_multilabel, thresholds=thresholds
        )
        if self.is_multilabel:
            return [
                [self.idx2class(i) for i in prediction] for prediction in predictions
            ]
        else:
            return [self.idx2class(i) for i in predictions]

    def score(
        self,
        X: Sequence[str] = None,
        y: Union[Sequence[Sequence[str]], Sequence[str]] = None,
        *,
        dataset: ClassificationDataset = None,
        thresholds: Union[float, List[float]] = 0.5,
        batch_size: int = 128
    ) -> pd.DataFrame:
        dataset = self.prepare_dataset(X, y, dataset)
        predictions = self.predict(
            dataset=dataset, thresholds=thresholds, batch_size=batch_size
        )
        return classification_fscore(dataset.y, predictions, classes=self.classes)

    @classmethod
    def format_score(cls, score_df: pd.DataFrame, format: str = "markdown") -> str:
        return tabulate(score_df, headers="keys", tablefmt="github", showindex=False)

    def get_config(self) -> Dict[str, Any]:
        return {
            **super().get_config(),
            "is_multilabel": self.is_multilabel,
            "use_batch_normalization": self.use_batch_normalization,
        }

    @classmethod
    def _filter_config(cls, config):
        config = super()._filter_config(config)
        config.pop("algorithm", None)
        config.pop("task", None)
        return config

    @classmethod
    def get_custom_objects(cls) -> Dict[str, Any]:
        return {
            **super().get_custom_objects(),
            "PrecisionWithLogits": PrecisionWithLogits,
            "RecallWithLogits": RecallWithLogits,
            "FBetaScoreWithLogits": FBetaScoreWithLogits,
        }


from typing import Optional, Dict, Any

import tensorflow as tf
from tensorflow.keras.layers import Embedding

from sknlp.vocab import Vocab
from sknlp.typing import WeightRegularizer, WeightInitializer, WeightConstraint

from .text2vec import Text2vec


class Word2vec(Text2vec):
    def __init__(
        self,
        vocab: Vocab,
        embedding_size: int,
        segmenter: str = "jieba",
        max_sequence_length: Optional[int] = None,
        sequence_length: Optional[int] = None,
        embeddings_initializer: WeightInitializer = "uniform",
        embeddings_regularizer: Optional[WeightRegularizer] = None,
        embeddings_constraint: Optional[WeightConstraint] = None,
        name: str = "word2vec",
    ) -> None:
        """
        基础符号->向量模块.

        Parameters:
        ----------
        vocab: sknlp.vocab.Vocab. 符号表.
        embed_size: int > 0. 向量维度.
        name: str. 模块名.
        embeddings_initializer: Initializer for the `embeddings` matrix.
        embeddings_regularizer: Regularizer function applied to
        the `embeddings` matrix.
        embeddings_constraint: Constraint function applied to
        the `embeddings` matrix.
        input_length: Length of input sequences, when it is constant.
        This argument is required if you are going to connect
        `Flatten` then `Dense` layers upstream
        (without it, the shape of the dense outputs cannot be computed).

        Inputs
        ----------
        2D tensor with shape: `(batch_size, input_length)`.

        Outputs
        ----------
        3D tensor with shape: `(batch_size, input_length, embed_size)`.
        """
        super().__init__(
            vocab,
            segmenter=segmenter,
            max_sequence_length=max_sequence_length,
            sequence_length=sequence_length,
            name=name,
        )
        self._embedding_size = embedding_size
        embedding = Embedding(
            len(vocab),
            embedding_size,
            embeddings_initializer=embeddings_initializer,
            embeddings_regularizer=embeddings_regularizer,
            embeddings_constraint=embeddings_constraint,
            mask_zero=True,
            name="embeddings",
        )
        self._model = tf.keras.Sequential(embedding, name=name)


from typing import Optional, Dict, Any, Callable
import os
import tempfile
from enum import Enum
from collections import Counter

import tensorflow as tf
from tensorflow.keras.initializers import TruncatedNormal
from official.modeling import activations
from official.nlp.bert.configs import BertConfig

from sknlp.vocab import Vocab
from sknlp.layers import BertLayer
from sknlp.layers.utils import (
    convert_bert_checkpoint,
    convert_electra_checkpoint,
    convert_albert_checkpoint,
)
from .text2vec import Text2vec


class BertFamily(Enum):
    BERT = 1
    ALBERT = 2
    ELECTRA = 3


def get_activation(activation_string: str) -> Callable:
    if activation_string == "gelu":
        return activations.gelu
    else:
        return tf.keras.activations.get(activation_string)


class Bert2vec(Text2vec):
    def __init__(
        self,
        vocab: Vocab,
        segmenter: Optional[str] = None,
        hidden_size: int = 768,
        embedding_size: int = 768,
        num_layers: int = 12,
        num_attention_heads: int = 12,
        sequence_length: Optional[int] = None,
        max_sequence_length: int = 512,
        type_vocab_size: int = 16,
        intermediate_size: int = 3072,
        activation: activations = activations.gelu,
        dropout_rate: float = 0.1,
        attention_dropout_rate: float = 0.1,
        initializer: tf.keras.initializers.Initializer = TruncatedNormal(stddev=0.02),
        return_all_encoder_outputs: bool = False,
        share_layer: bool = False,
        cls_pooling: bool = True,
        name: str = "bert2vec",
        **kwargs
    ) -> None:
        super().__init__(
            vocab,
            segmenter=segmenter,
            sequence_length=sequence_length,
            max_sequence_length=max_sequence_length,
            name=name,
            **kwargs
        )
        if embedding_size is None:
            embedding_size = hidden_size
        self._embedding_size = embedding_size
        bert_layer = BertLayer(
            len(self.vocab.sorted_tokens),
            embedding_size=embedding_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            num_attention_heads=num_attention_heads,
            sequence_length=sequence_length,
            max_sequence_length=max_sequence_length,
            type_vocab_size=type_vocab_size,
            intermediate_size=intermediate_size,
            activation=activation,
            dropout_rate=dropout_rate,
            attention_dropout_rate=attention_dropout_rate,
            initializer=initializer,
            return_all_encoder_outputs=return_all_encoder_outputs,
            share_layer=share_layer,
            cls_pooling=cls_pooling,
            name=name,
        )
        token_ids = tf.keras.Input(
            shape=(sequence_length,), dtype=tf.int64, name="input_token_ids"
        )
        type_ids = tf.keras.Input(
            shape=(sequence_length,), dtype=tf.int64, name="input_type_ids"
        )
        outputs = bert_layer([token_ids, type_ids])
        self._model = tf.keras.Model(
            inputs=[token_ids, type_ids], outputs=outputs, name=name
        )

    @classmethod
    def from_tfv1_checkpoint(
        cls,
        model_type: BertFamily,
        v1_checkpoint: str,
        config_filename: str = "bert_config.json",
        cls_pooling: bool = True,
        sequence_length: Optional[int] = None,
        name: str = "bert",
    ):
        config = BertConfig.from_json_file(os.path.join(v1_checkpoint, config_filename))
        with open(os.path.join(v1_checkpoint, "vocab.txt")) as f:
            token_list = f.read().strip("\n").split("\n")
            vocab = Vocab(
                Counter(token_list), pad_token=token_list[0], unk_token=token_list[1]
            )
        share_layer, cls_pooling = False, True
        convert_checkpoint = convert_bert_checkpoint
        if model_type is BertFamily.ALBERT:
            share_layer = True
            convert_checkpoint = convert_albert_checkpoint
        if model_type is BertFamily.ELECTRA:
            cls_pooling = False
            convert_checkpoint = convert_electra_checkpoint
        activation = get_activation(config.hidden_act)
        module = cls(
            vocab,
            embedding_size=config.embedding_size,
            hidden_size=config.hidden_size,
            num_layers=config.num_hidden_layers,
            num_attention_heads=config.num_attention_heads,
            intermediate_size=config.intermediate_size,
            activation=activation,
            dropout_rate=config.hidden_dropout_prob,
            attention_dropout_rate=config.attention_probs_dropout_prob,
            sequence_length=sequence_length,
            max_sequence_length=config.max_position_embeddings,
            type_vocab_size=config.type_vocab_size,
            initializer=tf.keras.initializers.TruncatedNormal(
                stddev=config.initializer_range
            ),
            share_layer=share_layer,
            cls_pooling=cls_pooling,
            name=name,
        )
        with tempfile.TemporaryDirectory() as temp_dir:
            temporary_checkpoint = os.path.join(temp_dir, "ckpt")
            convert_checkpoint(
                checkpoint_from_path=v1_checkpoint,
                checkpoint_to_path=temporary_checkpoint,
                num_heads=config.num_attention_heads,
                converted_root_name=name,
            )
            module._model.load_weights(
                temporary_checkpoint
            ).assert_existing_objects_matched()
        return module

    def export(
        self,
        directory: str,
        name: str,
        version: str = "0",
        only_output_cls: bool = False,
    ) -> None:
        if only_output_cls:
            self._model = tf.keras.Model(
                inputs=self._model.inputs,
                outputs=self._model.outputs[1:],
                name="bert2vec",
            )
        super().export(directory, name, version)
        d = os.path.join(directory, name, version)
        self.save_vocab(d)

    def get_custom_objects(self) -> Dict[str, Any]:
        return {**super().get_custom_objects(), "BertLayer": BertLayer}


from typing import Sequence, List, Dict, Any, Optional
from functools import partial

import pandas as pd
from tabulate import tabulate
import tensorflow as tf

from sknlp.vocab import Vocab
from sknlp.data import TaggingDataset
from sknlp.callbacks import ModelScoreCallback
from sknlp.layers import CrfDecodeLayer

from ..supervised_model import SupervisedNLPModel
from ..text2vec import Text2vec

from .utils import tagging_fscore, viterbi_decode


class DeepTagger(SupervisedNLPModel):
    def __init__(
        self,
        classes: Sequence[str],
        start_tag: Optional[str] = None,
        end_tag: Optional[str] = None,
        use_crf: bool = True,
        max_sequence_length: Optional[int] = None,
        sequence_length: Optional[int] = None,
        segmenter: str = "char",
        embedding_size: int = 100,
        use_batch_normalization: bool = True,
        text2vec: Optional[Text2vec] = None,
        **kwargs,
    ):
        classes = list(classes)
        self._start_tag = start_tag
        self._end_tag = end_tag
        if start_tag is not None and end_tag is not None:
            if start_tag not in classes:
                classes.append(start_tag)
            if end_tag not in classes:
                classes.append(end_tag)
        self._pad_tag = "[PAD]"
        if self._pad_tag != classes[0]:
            classes.insert(0, self._pad_tag)
        self._use_crf = use_crf
        self._use_batch_normalization = use_batch_normalization
        super().__init__(
            classes,
            max_sequence_length=max_sequence_length,
            sequence_length=sequence_length,
            segmenter=segmenter,
            embedding_size=embedding_size,
            text2vec=text2vec,
            task="tagging",
            **kwargs,
        )

    @property
    def use_crf(self) -> bool:
        return self._use_crf

    @property
    def use_batch_normalization(self) -> bool:
        return self._use_batch_normalization

    @property
    def start_tag(self) -> str:
        return self._start_tag

    @property
    def end_tag(self) -> str:
        return self._end_tag

    @property
    def pad_tag(self) -> str:
        return self._pad_tag

    def get_loss(self) -> None:
        return None

    def get_callbacks(self, *args, **kwargs) -> List[tf.keras.callbacks.Callback]:
        callbacks = super().get_callbacks(*args, **kwargs)
        if self.valid_dataset is not None:
            score_func = partial(
                self.score,
                dataset=self.valid_dataset,
                batch_size=kwargs.get("batch_size", None) or 128,
            )
            callbacks.append(ModelScoreCallback(score_func))
        return callbacks

    def get_metrics(self) -> List[tf.keras.metrics.Metric]:
        return []

    @classmethod
    def get_monitor(cls) -> str:
        return "val_fscore"

    def create_dataset_from_df(
        self,
        df: pd.DataFrame,
        vocab: Vocab,
        segmenter: str,
        labels: Sequence[str],
        no_label: bool,
    ) -> TaggingDataset:
        return TaggingDataset(
            vocab,
            list(labels),
            df=df,
            max_length=self.max_sequence_length,
            no_label=no_label,
            text_segmenter=segmenter,
            start_tag=self.start_tag,
            end_tag=self.end_tag,
        )

    def predict(
        self,
        X: Sequence[str] = None,
        *,
        dataset: TaggingDataset = None,
        batch_size: int = 128,
    ) -> List[List[str]]:
        emissions, mask = super().predict(X, dataset=dataset, batch_size=batch_size)
        tag_ids_list = viterbi_decode(
            self._model.get_layer("crf").get_weights()[0],
            emissions.transpose(1, 0, 2),
            mask.transpose(1, 0),
        )
        exclude_tags = {self.pad_tag, self.start_tag, self.end_tag}
        predictions = []
        for tag_ids in tag_ids_list:
            tags = []
            for tag_id in tag_ids:
                tag = self.idx2class(tag_id)
                if tag not in exclude_tags:
                    tags.append(tag)
                if tag == self.pad_tag:
                    break
            predictions.append(tags)
        return predictions

    def score(
        self,
        X: Sequence[str] = None,
        y: Sequence[Sequence[str]] = None,
        *,
        dataset: TaggingDataset = None,
        batch_size: int = 128,
    ) -> pd.DataFrame:
        dataset = self.prepare_dataset(X, y, dataset)
        predictions = self.predict(dataset=dataset, batch_size=batch_size)
        labels = list(set([c.split("-")[-1] for c in self.classes if "-" in c]))
        return tagging_fscore(dataset.X, dataset.y, predictions, labels)

    @classmethod
    def format_score(self, score_df: pd.DataFrame, format: str = "markdown") -> str:
        return tabulate(score_df, headers="keys", tablefmt="github", showindex=False)

    def export(self, directory: str, name: str, version: str = "0") -> None:
        if self.use_crf:
            mask = self._model.get_layer("mask_layer").output
            emissions = self._model.get_layer("mlp").output
            crf = CrfDecodeLayer(self.num_classes, self.max_sequence_length)
            crf.set_weights(self._model.get_layer("crf").get_weights())
            model = tf.keras.Model(
                inputs=self._model.inputs[0], outputs=crf(emissions, mask)
            )
            original_model = self._model
            self._model = model
            super().export(directory, name, version=version)
            self._model = original_model
        else:
            super().export(directory, name, version=version)

    def get_config(self) -> Dict[str, Any]:
        return {
            **super().get_config(),
            "start_tag": self.start_tag,
            "end_tag": self.end_tag,
            "use_crf": self.use_crf,
            "use_batch_normalization": self.use_batch_normalization,
        }

    @classmethod
    def _filter_config(cls, config: Dict[str, Any]) -> Dict[str, Any]:
        config = super()._filter_config(config)
        config.pop("algorithm", None)
        config.pop("task", None)
        return config

    @classmethod
    def get_custom_objects(cls) -> Dict[str, Any]:
        return {
            **super().get_custom_objects(),
            "CrfDecodeLayer": CrfDecodeLayer,
            "Orthogonal": tf.keras.initializers.Orthogonal,
        }


from typing import Callable, Optional, List, Union, Sequence, Tuple, Any
import string

import pandas as pd
import tensorflow as tf
from .text_segmenter import get_segmenter


class NLPDataset:
    def __init__(
        self,
        df: Optional[pd.DataFrame] = None,
        csv_file: Optional[str] = None,
        in_memory: bool = True,
        no_label: bool = False,
        text_segmenter: str = "char",
        max_length: Optional[int] = None,
        na_value: str = "",
        column_dtypes: List[str] = ["str", "str"],
        text_dtype: tf.DType = tf.string,
        label_dtype: tf.DType = tf.string,
        batch_padding_shapes: Optional[Tuple[tf.DType]] = None,
        batch_padding_values: Optional[Tuple[tf.DType]] = None,
    ):
        assert (
            df is not None or csv_file is not None
        ), "either df or csv_file may not be None"
        self.in_memory = in_memory
        self.na_value = na_value
        self.column_dtypes = column_dtypes
        if df is not None:
            self._original_dataset = self.dataframe_to_dataset(df, na_value)
            self.size = df.shape[0]
        elif csv_file is not None:
            self._original_dataset, self.size = self.load_csv(
                csv_file, "\t", in_memory, column_dtypes, na_value
            )
        self.no_label = no_label
        self.text_cutter = get_segmenter(text_segmenter)
        self.max_length = max_length or 99999
        self.text_dtype = text_dtype
        self.label_dtype = label_dtype
        self.batch_padding_shapes = batch_padding_shapes
        self.batch_padding_values = batch_padding_values
        self._dataset = self._transform(self._original_dataset)

    @property
    def X(self) -> List[Any]:
        x = []
        for data in self._original_dataset.as_numpy_iterator():
            texts = data[:-1]
            if len(texts) == 1:
                x.append(texts[0].decode("utf-8"))
            else:
                x.append([text.decode("utf-8") for text in texts])
        return x

    @property
    def y(self) -> List[str]:
        if self.no_label:
            return []
        return [
            data[-1].decode("utf-8")
            for data in self._original_dataset.as_numpy_iterator()
        ]

    def _text_transform(self, text: tf.Tensor) -> Union[List[str], str]:
        return self.text_cutter(text.numpy().decode("utf-8"))[: self.max_length]

    def _label_transform(self, label: tf.Tensor) -> str:
        return label.numpy().decode("utf-8")

    def _transform(self, dataset: tf.data.Dataset) -> tf.data.Dataset:
        def func(*data):
            text, label = data
            return self._text_transform(text), self._label_transform(label)

        return dataset.map(
            lambda *data: tf.py_function(
                func,
                inp=data,
                Tout=(self.text_dtype,)
                if self.no_label
                else (self.text_dtype, self.label_dtype),
            ),
            num_parallel_calls=tf.data.experimental.AUTOTUNE,
        )

    def shuffled_dataset(
        self, shuffle_buffer_size: Optional[int] = None
    ) -> tf.data.Dataset:
        shuffle_buffer_size = shuffle_buffer_size or self.size or 100000
        return self._dataset.shuffle(shuffle_buffer_size)

    def batchify(
        self,
        batch_size: int,
        shuffle: bool = True,
        shuffle_buffer_size: Optional[int] = None,
        before_batch: Optional[Callable] = None,
        after_batch: Optional[Callable] = None,
    ) -> tf.data.Dataset:
        dataset = (
            self.shuffled_dataset(shuffle_buffer_size) if shuffle else self._dataset
        )
        if before_batch is not None:
            dataset = dataset.map(before_batch)
        if self.batch_padding_shapes is None or self.batch_padding_values is None:
            dataset = dataset.batch(batch_size)
        else:
            dataset = dataset.padded_batch(
                batch_size,
                padded_shapes=self.batch_padding_shapes,
                padding_values=self.batch_padding_values,
            )
        if after_batch is not None:
            dataset = dataset.map(after_batch)
        return dataset.prefetch(buffer_size=tf.data.experimental.AUTOTUNE)

    def dataframe_to_dataset(self, df: pd.DataFrame, na_value: str) -> tf.data.Dataset:
        df.fillna(na_value, inplace=True)
        return tf.data.Dataset.from_tensor_slices(tuple(df[col] for col in df.columns))

    def load_csv(
        self,
        filename: str,
        sep: str,
        in_memory: bool,
        column_dtypes: List[str],
        na_value: str,
    ) -> Tuple[tf.data.Dataset, Optional[int]]:
        if in_memory:
            df = pd.read_csv(filename, sep=sep, quoting=3)
            for dtype, col in zip(column_dtypes, df.columns):
                df[col] = df[col].astype(dtype)
            return self.dataframe_to_dataset(df, na_value), df.shape[0]
        tf_dtype_mapping = {"str": tf.string, "int": tf.int32, "float": tf.float32}
        return (
            tf.data.experimental.CsvDataset(
                filename,
                [
                    tf_dtype_mapping.get(dtype.rstrip(string.digits), "str")
                    for dtype in column_dtypes[: -1 if self.no_label else None]
                ],
                header=True,
                field_delim=sep,
                na_value=na_value,
            ),
            None,
        )



from typing import Sequence, List, Optional, Tuple

import numpy as np
import pandas as pd
import tensorflow as tf

from sknlp.vocab import Vocab
from .nlp_dataset import NLPDataset


def _combine(text, context, label):
    return (text, context), label


class SimilarityDataset(NLPDataset):
    def __init__(
        self,
        vocab: Vocab,
        labels: Sequence[str],
        df: Optional[pd.DataFrame] = None,
        csv_file: Optional[str] = None,
        in_memory: bool = True,
        no_label: bool = False,
        max_length: Optional[int] = None,
        text_segmenter: str = "char",
        text_dtype: tf.DType = tf.int32,
        label_dtype: tf.DType = tf.float32,
        batch_padding_shapes: Optional[Tuple[tf.DType]] = ((None,), (None,), (None,)),
        batch_padding_values: Optional[Tuple[tf.DType]] = (0, 0, 0.0),
    ):
        self.vocab = vocab
        self.label2idx = dict(zip(labels, range(len(labels))))
        super().__init__(
            df=df,
            csv_file=csv_file,
            in_memory=in_memory,
            no_label=no_label,
            text_segmenter=text_segmenter,
            max_length=max_length,
            na_value=0.0,
            column_dtypes=["str", "str", "float32"],
            text_dtype=text_dtype,
            label_dtype=label_dtype,
            batch_padding_shapes=batch_padding_shapes,
            batch_padding_values=batch_padding_values,
        )

    @property
    def y(self) -> List[float]:
        if self.no_label:
            return []
        return [float(data[-1]) for data in self._original_dataset.as_numpy_iterator()]

    def _text_transform(self, text: tf.Tensor) -> np.ndarray:
        tokens = super()._text_transform(text)
        return np.array(
            [self.vocab[t] for t in tokens[: self.max_length]], dtype=np.int32
        )

    def _label_transform(self, label: tf.Tensor) -> float:
        return label

    def _transform(self, dataset: tf.data.Dataset) -> tf.data.Dataset:
        def func(*data):
            text, context, label = data
            return (
                self._text_transform(text),
                self._text_transform(context),
                self._label_transform(label),
            )

        return dataset.map(
            lambda *data: tf.py_function(
                func,
                inp=data,
                Tout=(self.text_dtype, self.text_dtype)
                if self.no_label
                else (self.text_dtype, self.text_dtype, self.label_dtype),
            ),
            num_parallel_calls=tf.data.experimental.AUTOTUNE,
        )

    def batchify(
        self,
        batch_size: int,
        shuffle: bool = True,
        shuffle_buffer_size: Optional[int] = None,
    ) -> tf.data.Dataset:
        return super().batchify(
            batch_size,
            shuffle=shuffle,
            shuffle_buffer_size=shuffle_buffer_size,
            after_batch=_combine,
        )


from typing import Sequence, Optional, Union, Any, Dict
import functools

import tensorflow as tf
from tensorflow.keras.metrics import Precision, Recall
from tensorflow_addons.metrics import FBetaScore


class PrecisionWithLogits(Precision):
    def __init__(
        self,
        thresholds: Union[float, Sequence[float], None] = None,
        top_k: Optional[int] = None,
        class_id: Optional[int] = None,
        name: str = "precision",
        dtype: Optional[tf.DType] = None,
        logits2scores: str = "sigmoid",
    ) -> None:
        super().__init__(
            thresholds=thresholds,
            top_k=top_k,
            class_id=class_id,
            name=name,
            dtype=dtype,
        )
        self.logits2scores = logits2scores
        if logits2scores == "sigmoid":
            self._l2s = tf.math.sigmoid
        elif logits2scores == "softmax":
            self._l2s = functools.partial(tf.math.softmax, axis=-1)

    def update_state(
        self,
        y_true: tf.Tensor,
        y_logits: tf.Tensor,
        sample_weight: Optional[tf.Tensor] = None,
    ) -> None:
        super().update_state(y_true, self._l2s(y_logits), sample_weight=sample_weight)

    def get_config(self) -> Dict[str, Any]:
        return {**super().get_config(), "logits2scores": self.logits2scores}


class RecallWithLogits(Recall):
    def __init__(
        self,
        thresholds: Union[float, Sequence[float], None] = None,
        top_k: Optional[int] = None,
        class_id: Optional[int] = None,
        name: str = "recall",
        dtype: Optional[tf.DType] = None,
        logits2scores: str = "sigmoid",
    ) -> None:
        super().__init__(
            thresholds=thresholds,
            top_k=top_k,
            class_id=class_id,
            name=name,
            dtype=dtype,
        )
        self.logits2scores = logits2scores
        if logits2scores == "sigmoid":
            self._l2s = tf.math.sigmoid
        elif logits2scores == "softmax":
            self._l2s = functools.partial(tf.math.softmax, axis=-1)

    def update_state(
        self,
        y_true: tf.Tensor,
        y_logits: tf.Tensor,
        sample_weight: Optional[tf.Tensor] = None,
    ) -> None:
        super().update_state(y_true, self._l2s(y_logits), sample_weight=sample_weight)

    def get_config(self) -> Dict[str, Any]:
        return {**super().get_config(), "logits2scores": self.logits2scores}


class FBetaScoreWithLogits(FBetaScore):
    def __init__(
        self,
        num_classes: int,
        average: str = "micro",
        beta: float = 1.0,
        thresholds: float = 0.5,
        name: str = "fbeta_score",
        dtype: Optional[tf.DType] = None,
        logits2scores: str = "sigmoid",
    ) -> None:
        super().__init__(
            num_classes,
            average=average,
            beta=beta,
            threshold=thresholds,
            name=name,
            dtype=dtype,
        )
        self.logits2scores = logits2scores
        if logits2scores == "sigmoid":
            self._l2s = tf.math.sigmoid
        elif logits2scores == "softmax":
            self._l2s = functools.partial(tf.math.softmax, axis=-1)

    def update_state(
        self,
        y_true: tf.Tensor,
        y_logits: tf.Tensor,
        sample_weight: Optional[tf.Tensor] = None,
    ) -> None:
        super().update_state(y_true, self._l2s(y_logits), sample_weight=sample_weight)

    def get_config(self) -> Dict[str, Any]:
        configs = super().get_config()
        configs.pop("threshold", None)
        return {
            **configs,
            "logits2scores": self.logits2scores,
            "thresholds": self.threshold,
        }


import functools
import jieba_fast as jieba


def get_segmenter(name):
    if name == 'jieba':
        return functools.partial(jieba.lcut, HMM=False)
    elif name == 'list' or name == 'char':
        return list
    elif name is None:
        return lambda x: x
    else:
        raise ValueError('unknown segmenter %s' % name)



from typing import Optional, Dict, Any, List, Union

import json
import os
import shutil
import tempfile

import tensorflow as tf

from sknlp.vocab import Vocab
from sknlp.utils import make_tarball

from sknlp.module.base_model import BaseNLPModel


class Text2vec(BaseNLPModel):
    def __init__(
        self,
        vocab: Vocab,
        segmenter: Optional[str] = None,
        max_sequence_length: Optional[int] = None,
        sequence_length: Optional[int] = None,
        embedding_size: int = 100,
        name: str = "text2vec",
    ) -> None:
        """
        基础符号->向量模块.

        Parameters:
        ----------
        vocab: sknlp.vocab.Vocab. 符号表.
        embed_size: int > 0. 向量维度.
        name: str. 模块名.
        embeddings_initializer: Initializer for the `embeddings` matrix.
        embeddings_regularizer: Regularizer function applied to
        the `embeddings` matrix.
        embeddings_constraint: Constraint function applied to
        the `embeddings` matrix.
        input_length: Length of input sequences, when it is constant.
        This argument is required if you are going to connect
        `Flatten` then `Dense` layers upstream
        (without it, the shape of the dense outputs cannot be computed).

        Inputs
        ----------
        2D tensor with shape: `(batch_size, input_length)`.

        Outputs
        ----------
        3D tensor with shape: `(batch_size, input_length, embed_size)`.
        """
        self._vocab = vocab
        self._segmenter = segmenter
        super().__init__(
            max_sequence_length=max_sequence_length,
            sequence_length=sequence_length,
            name=name,
        )
        self._embedding_size = embedding_size
        self._model: tf.keras.Model = None

    def __call__(self, inputs: Union[tf.Tensor, List[tf.Tensor]]) -> tf.Tensor:
        return self._model(inputs)

    @property
    def embedding_size(self):
        return self._embedding_size

    @property
    def vocab(self) -> Vocab:
        return self._vocab

    @property
    def segmenter(self) -> str:
        return self._segmenter

    def save_vocab(self, directory: str, filename: str = "vocab.json") -> None:
        with open(os.path.join(directory, filename), "w") as f:
            f.write(self.vocab.to_json())

    def save(self, directory: str) -> None:
        super().save(directory)
        self.save_vocab(directory)

    @classmethod
    def load(cls, directory: str) -> "Text2vec":
        with open(os.path.join(directory, "meta.json")) as f:
            meta = json.loads(f.read())
        with open(os.path.join(directory, "vocab.json")) as f:
            vocab = Vocab.from_json(f.read())
        module = cls.from_config({"vocab": vocab, **meta})
        module._model = tf.keras.models.load_model(
            os.path.join(directory, "model"), custom_objects=module.get_custom_objects()
        )
        module._built = True
        return module

    def save_archive(self, filename: str) -> None:
        with tempfile.TemporaryDirectory() as temp_dir:
            self.save(temp_dir)
            if not filename.endswith(".tar"):
                filename = ".".join([filename, "tar"])
            make_tarball(filename, temp_dir)

    @classmethod
    def load_archive(cls, filename: str) -> "Text2vec":
        with tempfile.TemporaryDirectory() as temp_dir:
            shutil.unpack_archive(filename, temp_dir, format="tar")
            return cls.load(temp_dir)

    def export(self, directory: str, name: str, version: str = "0") -> None:
        super().export(directory, name, version)
        d = os.path.join(directory, name, version)
        self.save_vocab(d)

    def get_config(self) -> Dict[str, Any]:
        return {
            **super().get_config(),
            "segmenter": self.segmenter,
            "embedding_size": self.embedding_size,
        }

    def get_inputs(self) -> tf.Tensor:
        return self._model.input

    def get_outputs(self) -> tf.Tensor:
        return self._model.output

    @property
    def weights(self) -> List[tf.Tensor]:
        return self._model.get_weights()

    def set_weights(self, weights: List[tf.Tensor]) -> None:
        self._model.set_weights(weights)



import os
import tarfile


def make_tarball(filename, root_dir):
    """
    将`root_dir`下的文件打包为一个tar文件

    filename: 打包后的文件名, 如果结尾不为.tar, 打包后的文件名为`filename`.tar
    root_dir: 需要打包的文件所在的目录
    """
    if not filename.endswith('.tar'):
        filename = '.'.join([filename, 'tar'])
    with tarfile.open(filename, 'w') as f:
        for file in os.listdir(root_dir):
            f.add(os.path.join(root_dir, file), arcname=file)



from typing import Dict, Any, Optional, Callable
import logging
import tensorflow as tf


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream = logging.StreamHandler()
logger.addHandler(stream)


class ModelScoreCallback(tf.keras.callbacks.Callback):

    def __init__(self, score_func: Callable):
        super().__init__()
        self._score_func = score_func

    def on_epoch_end(self, epoch: int, logs: Optional[Dict[str, Any]] = None):
        score_df = self._score_func()
        row = score_df[score_df["class"] == "avg"]
        logger.info(row)
        for col in score_df.columns:
            if col == "support":
                continue
            logs[f"val_{col}"] = row[col].values.tolist()[0]



import numpy as np
import tensorflow as tf


class ModelSave(tf.keras.callbacks.Callback):

    def __init__(self, model, dataset):
        super().__init__()
        self.model = model
        self.dataset = dataset

    def on_epoch_end(self, epoch, logs=None):

        np.save('arr-%d' % epoch, self.model.predict(self.dataset))



from typing import Sequence, List, Optional

import numpy as np
import pandas as pd

import tensorflow as tf

from sknlp.vocab import Vocab

from .classification_dataset import ClassificationDataset
from .tagging_dataset import TaggingDataset
from .similarity_dataset import SimilarityDataset


class BertClassificationDataset(ClassificationDataset):
    def __init__(
        self,
        vocab: Vocab,
        labels: Sequence[str],
        df: Optional[pd.DataFrame] = None,
        csv_file: Optional[str] = None,
        in_memory: bool = True,
        no_label: bool = False,
        is_multilabel: bool = True,
        max_length: Optional[int] = None,
    ):
        super().__init__(
            vocab,
            labels,
            df=df,
            csv_file=csv_file,
            in_memory=in_memory,
            no_label=no_label,
            is_multilabel=is_multilabel,
            max_length=max_length,
            text_segmenter=None,
            text_dtype=tf.string,
            label_dtype=tf.float32,
            batch_padding_shapes=None,
            batch_padding_values=None,
        )

    def _text_transform(self, text: tf.Tensor) -> str:
        return text.numpy().decode("utf-8")[: self.max_length]


class BertTaggingDataset(TaggingDataset):
    def __init__(
        self,
        vocab: Vocab,
        labels: Sequence[str],
        df: Optional[pd.DataFrame] = None,
        csv_file: Optional[str] = None,
        in_memory: bool = True,
        no_label: bool = False,
        max_length: Optional[int] = None,
    ):
        super().__init__(
            vocab,
            labels,
            df=df,
            csv_file=csv_file,
            in_memory=in_memory,
            no_label=no_label,
            start_tag="[CLS]",
            end_tag="[SEP]",
            text_segmenter=None,
            max_length=max_length,
            text_dtype=tf.string,
            label_dtype=tf.int32,
            batch_padding_shapes=((), (None,)),
            batch_padding_values=(vocab.pad, 0),
        )

    def _text_transform(self, text: tf.Tensor) -> str:
        return text.numpy().decode("utf-8")[: self.max_length]


class BertSimilarityDataset(SimilarityDataset):
    def __init__(
        self,
        vocab: Vocab,
        labels: Sequence[str],
        df: Optional[pd.DataFrame] = None,
        csv_file: Optional[str] = None,
        in_memory: bool = True,
        no_label: bool = False,
        max_length: Optional[int] = None,
    ):
        super().__init__(
            vocab,
            labels,
            df=df,
            csv_file=csv_file,
            in_memory=in_memory,
            no_label=no_label,
            max_length=max_length,
            text_segmenter=None,
            text_dtype=tf.string,
            label_dtype=tf.float32,
            batch_padding_shapes=None,
            batch_padding_values=None,
        )

    def _text_transform(self, text: tf.Tensor) -> str:
        return text.numpy().decode("utf-8")[: self.max_length]


from .deep_classifier import DeepClassifier
from .rnn_classifier import TextRNNClassifier
from .rcnn_classifier import TextRCNNClassifier
from .bert_classifier import BertClassifier


__all__ = [
    "DeepClassifier",
    "TextRNNClassifier",
    "TextRCNNClassifier",
    "BertClassifier",
]



from typing import List, Optional, Callable

import tensorflow as tf

from .model_score import ModelScoreCallback
from .save import ModelSave
from .weight_decay import WeightDecayScheduler


def create_exponential_decay_scheduler(
    update_factor: float, update_epochs: int
) -> Callable[[int, float], float]:
    def scheduler(epoch, x):
        return x * update_factor ** ((epoch + 1) % update_epochs == 0)

    return scheduler


def default_supervised_model_callbacks(
    learning_rate_update_factor: float,
    learning_rate_update_epochs: int,
    use_weight_decay: bool = False,
    enable_early_stopping: bool = False,
    early_stopping_monitor: str = "val_loss",
    early_stopping_monitor_direction: str = "min",
    early_stopping_patience: Optional[int] = None,
    early_stopping_min_delta: float = 0.0,
    early_stopping_use_best_epoch: bool = False,
    log_file: str = None,
    verbose: int = 2,
) -> List[tf.keras.callbacks.Callback]:
    scheduler = create_exponential_decay_scheduler(
        learning_rate_update_factor, learning_rate_update_epochs
    )
    learning_rate_decay = tf.keras.callbacks.LearningRateScheduler(
        scheduler, verbose=verbose
    )
    callbacks = [learning_rate_decay]
    if use_weight_decay > 0:
        callbacks.append(WeightDecayScheduler(scheduler, verbose=verbose))

    if enable_early_stopping:
        callbacks.append(
            tf.keras.callbacks.EarlyStopping(
                monitor=early_stopping_monitor,
                min_delta=early_stopping_min_delta,
                patience=early_stopping_patience or learning_rate_update_epochs,
                mode=early_stopping_monitor_direction,
                restore_best_weights=early_stopping_use_best_epoch,
            )
        )
    if log_file is not None:
        callbacks.append(tf.keras.callbacks.CSVLogger(log_file))
    return callbacks


__all__ = [
    "ModelScoreCallback",
    "WeightDecayScheduler",
    "ModelSave",
    "default_supervised_model_callbacks",
]



from typing import Sequence, Optional, Dict, Any, Callable, List

import json
import os
import itertools
from collections import Counter

import tensorflow as tf

from sknlp.vocab import Vocab


class BaseNLPModel:
    def __init__(
        self,
        max_sequence_length: Optional[int] = None,
        sequence_length: Optional[int] = None,
        segmenter: Optional[str] = None,
        name: Optional[str] = None,
        **kwargs
    ) -> None:
        self._max_sequence_length = max_sequence_length
        self._sequence_length = sequence_length
        self._segmenter = segmenter
        self._name = name
        self._kwargs = kwargs
        self._model: tf.keras.Model = None
        self._built = False

    @property
    def max_sequence_length(self) -> Optional[int]:
        return self._max_sequence_length

    @property
    def sequence_length(self) -> Optional[int]:
        return self._sequence_length

    @property
    def segmenter(self) -> Optional[str]:
        return self._segmenter

    @staticmethod
    def build_vocab(
        texts: Sequence[str],
        segment_func: Callable[[str], Sequence[str]],
        min_frequency=5,
    ) -> Vocab:
        counter = Counter(
            itertools.chain.from_iterable(segment_func(text) for text in texts)
        )
        return Vocab(counter, min_frequency=min_frequency)

    def build_encode_layer(self, inputs: tf.Tensor) -> tf.Tensor:
        raise NotImplementedError()

    def build_output_layer(self, inputs: tf.Tensor) -> tf.Tensor:
        raise NotImplementedError()

    def build(self) -> None:
        if self._built:
            return
        self._model = tf.keras.Model(
            inputs=self.get_inputs(), outputs=self.get_outputs(), name=self._name
        )
        self._built = True

    def get_inputs(self) -> tf.Tensor:
        raise NotImplementedError()

    def get_outputs(self) -> tf.Tensor:
        raise NotImplementedError()

    def get_loss(self, *args, **kwargs) -> tf.keras.losses.Loss:
        raise NotImplementedError()

    def get_metrics(self, *args, **kwargs) -> List[tf.keras.metrics.Metric]:
        return []

    def get_callbacks(self, *args, **kwargs) -> List[tf.keras.callbacks.Callback]:
        return []

    @classmethod
    def get_monitor(cls) -> str:
        raise NotImplementedError()

    @classmethod
    def get_custom_objects(cls) -> Dict[str, Any]:
        return {}

    def freeze(self) -> None:
        for layer in self._model.layers:
            layer.trainable = False

    def save_config(self, directory: str, filename: str = "meta.json") -> None:
        with open(os.path.join(directory, filename), "w") as f:
            f.write(json.dumps(self.get_config(), ensure_ascii=False))

    def save(self, directory: str) -> None:
        self._model.save(os.path.join(directory, "model"), save_format="tf")
        self.save_config(directory)

    @classmethod
    def load(cls, directory: str) -> "BaseNLPModel":
        with open(os.path.join(directory, "meta.json")) as f:
            meta = json.loads(f.read())
        module = cls.from_config(meta)
        module._model = tf.keras.models.load_model(
            os.path.join(directory, "model"), custom_objects=module.get_custom_objects()
        )
        module._built = True
        return module

    def export(self, directory: str, name: str, version: str = "0") -> None:
        d = os.path.join(directory, name, version)

        model: tf.keras.Model = tf.keras.models.model_from_json(
            self._model.to_json(),
            custom_objects={
                **self.get_custom_objects(),
                "TruncatedNormal": tf.keras.initializers.TruncatedNormal,
            },
        )
        model.set_weights(self._model.get_weights())
        model.save(d, include_optimizer=False, save_format="tf")
        self.save_config(d)

    def get_config(self) -> Dict[str, Any]:
        return {
            "max_sequence_length": self.max_sequence_length,
            "sequence_length": self.sequence_length,
            "segmenter": self.segmenter,
            "name": self._name,
        }

    @classmethod
    def _filter_config(cls, config: Dict[str, Any]) -> Dict[str, Any]:
        return config

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "BaseNLPModel":
        return cls(**cls._filter_config(config))



from typing import List, Tuple, Dict, Union, Optional
import numpy as np
from sklearn.metrics import precision_recall_fscore_support
from scipy.special import expit, softmax


# (精准率, 召回率, F值, 数量)
FscoreTuple = Tuple[float, float, float, Optional[int]]


def logits2scores(logits: np.ndarray, is_multilabel: bool) -> np.ndarray:
    """
    logits转换为sigmoid或softmax分值.

    Parameters
    ----------
    logits: 输入的`logits`, shape(`n_samples`, `n_classes`)
    is_multilabel: 是否为多标签分类, 多标签分值使用sigmoid, 非多标签使用softmax

    Returns
    ----------
    返回一个shape(`n_samples`, `n_classes`)的``ndarray``.
    """
    if logits.ndim != 2:
        raise ValueError(
            'logits should have shape(`n_samples`, `n_classes`),'
            'but shape%s was given' % str(logits.shape)
        )
    if is_multilabel:
        return expit(logits)
    else:
        return softmax(logits, axis=1)


def scores2classes(
    scores: np.ndarray, is_multilabel: bool,
    threshold: Union[np.ndarray, float] = 0.5
) -> Union[List[int], List[List[int]]]:
    """
    根据给定的`threshold`, 将分类问题的`scores`解析为对应的类别.

    Parameters
    ----------
    scores: 输入的`logits`, shape(`n_samples`, `n_classes`)
    is_multilabel: 是否为多标签分类
    threshold: 正例判断的阈值(仅在多标签和二分类问题时有作用)

    Returns
    ----------
    返回一个长度为`n_samples`的``list``.

    如果是多标签分类, 每一个sample对应的结果为一个``list``,
    其中的每个``int``值为这个sample对应的类别.

    如果不是多标签分类, 则每一个sample对应的结果为一个``int``,
    表示这个sample对应的类别.
    """
    assert scores.ndim == 2
    if is_multilabel:
        return [
            np.where(scores[i, :] > threshold)[0].tolist()
            for i in range(scores.shape[0])
        ]
    else:
        return np.argmax(scores, axis=1).tolist()


def logits2classes(
    logits: np.ndarray, is_multilabel: bool,
    threshold: Union[np.ndarray, float] = 0.5
) -> Union[List[int], List[List[int]]]:
    """
    根据给定的`threshold`, 将分类问题的`logits`解析为对应的类别.

    Parameters
    ----------
    logits: 输入的`logits`, shape(`n_samples`, `n_classes`)
    is_multilabel: 是否为多标签分类
    threshold: 正例判断的阈值(仅在多标签和二分类问题时有作用)

    Returns
    ----------
    返回一个长度为`n_samples`的``list``.

    如果是多标签分类, 每一个sample对应的结果为一个``list``,
    其中的每个``int``值为这个sample对应的类别.

    如果不是多标签分类, 则每一个sample对应的结果为一个``int``,
    表示这个sample对应的类别.
    """
    assert logits.ndim == 2
    num_classes = logits.shape[1]
    if is_multilabel:
        return [
            np.where(expit(logits[i, :]) > threshold)[0].tolist()
            for i in range(logits.shape[0])
        ]
    elif num_classes == 2:
        return np.where(expit(logits[:, 1]) > threshold, 1, 0).tolist()
    else:
        return np.argmax(logits, axis=1).tolist()


def classify_f_score(
    y: List[List[int]], p: List[List[int]], is_multilabel: bool,
    beta: float = 1, labels: Optional[List[str]] = None
) -> Dict[Union[str, int], FscoreTuple]:
    """
    计算分类结果的F值.

    Parameters
    ----------
    y: 标注标签
    p: 预测标签
    is_multilabel: 是否是多标签分类
    beta: 加权值, 默认为1, f = (1 + beta**2) * (P * R) / (beta**2 * P + R)
    labels: 标签名, 如果不提供则取0...n_classes - 1

    Returns
    ----------
    返回一个dict.

    如果是非多标签的二分类问题, 则仅包含score一个key, value是一个Tuple
    是正例的(precision, recall, f score, num samples).

    其他情况下, key是一个类别或者avg, value是一个Tuple,
    是对应key的(precision, recall, f score, num samples).
    """
    scores: Dict[Union[str, int], FscoreTuple] = dict()
    _y, _p = np.array(y), np.array(p)
    assert _y.shape == _p.shape, 'y and p must have same shape'

    max_class_idx = _y.max()
    if max_class_idx <= 1 and not is_multilabel:
        score = precision_recall_fscore_support(_y, _p, average='binary')
        scores['score'] = score
    else:
        num_classes = _y.shape[1] if _y.ndim == 2 else max_class_idx + 1
        idx = list(range(num_classes))
        detail_score = precision_recall_fscore_support(_y, _p,
                                                       beta=beta, labels=idx)
        micro_score = precision_recall_fscore_support(
            _y, _p, beta=beta, labels=idx, average='micro'
        )
        for i, score in enumerate(zip(*detail_score)):
            if labels is None:
                scores[i] = score
            else:
                scores[labels[i]] = score
        scores['avg'] = micro_score
    return scores



from typing import Dict, Any, Callable

import tensorflow as tf
from tensorflow.keras.layers import Layer, Dense, BatchNormalization, InputSpec
from official.modeling import activations


def get_activation(activation_string: str) -> Callable:
    if activation_string == "gelu":
        return activations.gelu
    else:
        return tf.keras.activations.get(activation_string)


class MLPLayer(Layer):
    def __init__(
        self,
        num_layers: int,
        hidden_size: int = 256,
        output_size: int = 1,
        activation: str = "tanh",
        batch_normalization: bool = True,
        momentum: float = 0.9,
        epsilon: float = 1e-5,
        name: str = "mlp",
        **kwargs
    ) -> None:
        super().__init__(name=name, **kwargs)
        self.num_layers = num_layers
        self.hidden_size = hidden_size
        self.output_size = output_size
        self.activation = activation
        self.batch_normalization = batch_normalization
        self.momentum = momentum
        self.epsilon = epsilon
        self.input_spec = InputSpec(min_ndim=2)

        self.dense_layers = []
        self.batchnorm_layers = []
        for i in range(num_layers):
            if i == num_layers - 1:
                self.dense_layers.append(Dense(output_size, name="dense-%d" % i))
            else:
                self.dense_layers.append(
                    Dense(
                        hidden_size,
                        activation=get_activation(activation),
                        name="dense-%d" % i,
                    )
                )
                if self.batch_normalization:
                    self.batchnorm_layers.append(
                        BatchNormalization(axis=-1, momentum=momentum, epsilon=epsilon)
                    )

    def build(self, input_shape: tf.TensorShape) -> None:
        last_dim = input_shape[-1]
        self.input_spec = InputSpec(min_ndim=2, axes={-1: last_dim})
        super().build(input_shape)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        outputs = inputs
        if self.batch_normalization:
            for dense, batchnorm in zip(self.dense_layers[:-1], self.batchnorm_layers):
                outputs = batchnorm(dense(outputs))
        else:
            for dense in self.dense_layers[:-1]:
                outputs = dense(outputs)
        return self.dense_layers[-1](outputs)

    def compute_output_shape(self, input_shape: tf.TensorShape) -> tf.TensorShape:
        input_shape = input_shape.with_rank_at_least(2)
        if input_shape[-1] is None:
            raise ValueError(
                "The innermost dimension of input_shape must be defined, "
                "but saw: %s" % input_shape
            )
        return input_shape[:-1].concatenate(self.output_size)

    def get_config(self) -> Dict[str, Any]:
        config = {
            "num_layers": self.num_layers,
            "hidden_size": self.hidden_size,
            "output_size": self.output_size,
            "activation": self.activation,
            "batch_normalization": self.batch_normalization,
            "momentum": self.momentum,
            "epsilon": self.epsilon,
        }
        base_config = super().get_config()
        return dict(**base_config, **config)



from typing import Sequence, Union, Optional, Dict, Any, List, Tuple, Type

import os
import logging

import tensorflow as tf
import tensorflow_addons as tfa
import numpy as np
import pandas as pd

from sknlp.data import NLPDataset
from sknlp.data.text_segmenter import get_segmenter
from sknlp.vocab import Vocab
from sknlp.callbacks import default_supervised_model_callbacks
from .text2vec import Word2vec, Text2vec
from .base_model import BaseNLPModel

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream = logging.StreamHandler()
logger.addHandler(stream)


class SupervisedNLPModel(BaseNLPModel):
    def __init__(
        self,
        classes: List[str],
        max_sequence_length: Optional[int] = None,
        sequence_length: Optional[int] = None,
        segmenter: str = "jieba",
        embedding_size: int = 100,
        text2vec: Optional[Text2vec] = None,
        task: Optional[str] = None,
        algorithm: Optional[str] = None,
        **kwargs
    ) -> None:
        super().__init__(
            max_sequence_length=max_sequence_length,
            sequence_length=sequence_length,
            segmenter=text2vec.segmenter if text2vec else segmenter,
        )
        self._text2vec = text2vec
        if text2vec is not None:
            self._embedding_size = text2vec.embedding_size
            self._max_sequence_length = (
                max_sequence_length or text2vec.max_sequence_length
            )
            self._sequence_length = text2vec.sequence_length
        else:
            self._embedding_size = embedding_size
            self._max_sequence_length = max_sequence_length
            self._sequence_length = sequence_length

        self._task = task
        self._algorithm = algorithm
        self._class2idx = dict(zip(list(classes), range(len(classes))))
        self._idx2class = dict(zip(range(len(classes)), list(classes)))

    @property
    def embedding_size(self) -> int:
        return self._embedding_size

    @property
    def num_classes(self) -> int:
        return len(self._class2idx)

    @property
    def classes(self) -> List[str]:
        return list(self._class2idx.keys())

    @property
    def text2vec(self) -> Text2vec:
        return self._text2vec

    @property
    def train_dataset(self) -> NLPDataset:
        return getattr(self, "_train_dataset", None)

    @property
    def valid_dataset(self) -> NLPDataset:
        return getattr(self, "_valid_dataset", None)

    def class2idx(self, class_name: str) -> int:
        return self._class2idx.get(class_name, None)

    def idx2class(self, class_idx: int) -> str:
        return self._idx2class.get(class_idx, None)

    def get_inputs(self) -> tf.Tensor:
        if getattr(self, "inputs", None) is None:
            return self.text2vec.get_inputs()
        else:
            return self.inputs

    def get_outputs(self) -> tf.Tensor:
        return self.build_output_layer(self.build_encode_layer(self.get_inputs()))

    def build(self) -> None:
        if self._built:
            return
        assert self._text2vec is not None
        super().build()

    def create_dataset_from_df(
        self,
        df: pd.DataFrame,
        vocab: Vocab,
        segmenter: str,
        labels: Sequence[str],
        no_label: bool,
    ) -> NLPDataset:
        raise NotImplementedError()

    def prepare_dataset(
        self,
        X: Sequence[str],
        y: Union[Sequence[Sequence[str]], Sequence[str], Sequence[float]],
        dataset: NLPDataset,
    ) -> NLPDataset:
        assert X is not None or dataset is not None
        if dataset is not None:
            return dataset
        if (
            y is not None
            and isinstance(y[0], (list, tuple))
            and isinstance(y[0][0], str)
        ):
            y = ["|".join(map(str, y_i)) for y_i in y]
        if isinstance(X[0], (list, tuple)):
            df = pd.DataFrame(zip(*X, y) if y is not None else X)
        else:
            df = pd.DataFrame(zip(X, y) if y is not None else X)
        return self.create_dataset_from_df(
            df, self.text2vec.vocab, self.text2vec.segmenter, self.classes, y is None
        )

    def compile_optimizer(self, optimizer_name, **kwargs) -> None:
        if not self._built:
            self.build()
        weight_decay = kwargs.pop("weight_decay", 0)
        if optimizer_name != "adam":
            optimizer = tf.keras.optimizers.deserialize(
                {"class_name": optimizer_name, "config": kwargs}
            )
        elif weight_decay > 0:
            optimizer = tfa.optimizers.AdamW(weight_decay, **kwargs)
        else:
            optimizer = tfa.optimizers.LazyAdam(**kwargs)
        self._model.compile(
            optimizer=optimizer,
            loss=self.get_loss(),
            metrics=self.get_metrics(),
        )

    def fit(
        self,
        X: Sequence[str] = None,
        y: Union[Sequence[Sequence[str]], Sequence[str], Sequence[float]] = None,
        *,
        dataset: NLPDataset = None,
        valid_X: Sequence[str] = None,
        valid_y: Union[Sequence[Sequence[str]], Sequence[str], Sequence[float]] = None,
        valid_dataset: NLPDataset = None,
        batch_size: int = 128,
        n_epochs: int = 10,
        optimizer: str = "adam",
        optimizer_kwargs: Optional[Dict[str, Any]] = None,
        learning_rate: float = 1e-3,
        weight_decay: float = 0.0,
        clip: Optional[float] = 5.0,
        learning_rate_update_factor: float = 0.5,
        learning_rate_update_epochs: int = 10,
        enable_early_stopping: bool = False,
        early_stopping_patience: Optional[int] = None,
        early_stopping_min_delta: float = 0.0,
        early_stopping_use_best_epoch: bool = False,
        early_stopping_monitor: int = 2,  # 1 for loss, 2 for metric
        checkpoint: Optional[str] = None,
        save_frequency: int = 1,
        log_file: Optional[str] = None,
        verbose: int = 2
    ) -> None:
        if self._text2vec is None:
            cut = get_segmenter(self.segmenter)
            vocab = self.build_vocab(X or dataset.text, cut)
            self._text2vec = Word2vec(vocab, self.embedding_size, self.segmenter)

        self._train_dataset = self.prepare_dataset(X, y, dataset)
        if (valid_X is None or valid_y is None) and valid_dataset is None:
            self._valid_dataset = None
        else:
            self._valid_dataset = self.prepare_dataset(valid_X, valid_y, valid_dataset)
        train_tf_dataset = self.train_dataset.batchify(batch_size)
        valid_tf_dataset = None
        if self.valid_dataset is not None:
            valid_tf_dataset = self.valid_dataset.batchify(batch_size, shuffle=False)

        optimizer_kwargs = optimizer_kwargs or dict()
        optimizer_kwargs = {
            "learning_rate": learning_rate,
            "weight_decay": weight_decay,
            **optimizer_kwargs,
        }
        if clip is not None:
            optimizer_kwargs["clipnorm"] = clip
        self.compile_optimizer(optimizer, **optimizer_kwargs)

        monitor = "val_loss"
        monitor_direction = "min"
        if early_stopping_monitor == 2 and self.get_monitor():
            monitor = self.get_monitor()
            monitor_direction = "max"
        callbacks = default_supervised_model_callbacks(
            learning_rate_update_factor=learning_rate_update_factor,
            learning_rate_update_epochs=learning_rate_update_epochs,
            use_weight_decay=weight_decay > 0,
            enable_early_stopping=enable_early_stopping,
            early_stopping_monitor=monitor,
            early_stopping_monitor_direction=monitor_direction,
            early_stopping_patience=early_stopping_patience,
            early_stopping_min_delta=early_stopping_min_delta,
            early_stopping_use_best_epoch=early_stopping_use_best_epoch,
            log_file=log_file,
        )
        callbacks.extend(self.get_callbacks(batch_size))

        self._model.fit(
            train_tf_dataset,
            epochs=n_epochs,
            validation_data=valid_tf_dataset if self.valid_dataset else None,
            callbacks=callbacks,
            verbose=verbose,
        )

    def predict(
        self,
        X: Sequence[str] = None,
        *,
        dataset: NLPDataset = None,
        batch_size: int = 128
    ) -> np.ndarray:
        assert self._built
        dataset = self.prepare_dataset(X, None, dataset)
        return self._model.predict(dataset.batchify(batch_size, shuffle=False))

    def score(
        self,
        X: Sequence[str] = None,
        y: Union[Sequence[Sequence[str]], Sequence[str], Sequence[float]] = None,
        *,
        thresholds: Union[float, Dict[str, float]] = 0.5,
        batch_size: int = 128
    ) -> pd.DataFrame:
        raise NotImplementedError()

    def get_config(self) -> Dict[str, Any]:
        return {
            **super().get_config(),
            "classes": self.classes,
            "task": self._task,
            "algorithm": self._algorithm,
        }

    def save(self, directory: str) -> None:
        super().save(directory)
        self.text2vec.save_vocab(directory)

    @classmethod
    def load(cls, directory: str) -> "SupervisedNLPModel":
        module = super().load(directory)
        with open(os.path.join(directory, "vocab.json")) as f:
            vocab = Vocab.from_json(f.read())
        module._text2vec = Text2vec(
            vocab,
            segmenter=module.segmenter,
            max_sequence_length=module.max_sequence_length,
            sequence_length=module.sequence_length,
        )
        return module

    def export(self, directory: str, name: str, version: str = "0") -> None:
        super().export(directory, name, version)
        d = os.path.join(directory, name, version)
        self.text2vec.save_vocab(d)

    @classmethod
    def get_custom_objects(cls) -> Dict[str, Any]:
        return {**super().get_custom_objects(), "AdamW": tfa.optimizers.AdamW}



from typing import List, Optional, Sequence
import logging
from functools import partial

import numpy as np
import tensorflow as tf
import tensorflow.compat.v1 as tfv1


logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)
logger.addHandler(handler)
logger.setLevel(logging.INFO)


BERT_NAME_REPLACEMENTS = [
    ("embeddings_project", "embedding_projection"),
    ("encoder", "transformer"),
    ("embeddings/word_embeddings", "word_embeddings/embeddings"),
    ("embeddings/token_type_embeddings", "type_embeddings/embeddings"),
    ("embeddings/position_embeddings", "position_embedding/embeddings"),
    ("embeddings/LayerNorm", "embeddings/layer_norm"),
    ("attention/self", "self_attention"),
    ("attention/output/dense", "self_attention/attention_output"),
    ("attention/output/LayerNorm", "self_attention_layer_norm"),
    ("intermediate/dense", "intermediate"),
    ("output/dense", "output"),
    ("output/LayerNorm", "output_layer_norm"),
    ("pooler/dense", "pooler_transform"),
    ("cls/predictions/output_bias", "cls/predictions/output_bias/bias"),
    ("cls/seq_relationship/output_bias", "predictions/transform/logits/bias"),
    ("cls/seq_relationship/output_weights", "predictions/transform/logits/kernel"),
]
BERT_PERMUTATIONS = [
    ("cls/seq_relationship/output_weights", (1, 0)),
]


def _bert_name_replacement(var_name, name_replacements):
    """Gets the variable name replacement."""
    for src_pattern, tgt_pattern in name_replacements:
        if src_pattern in var_name:
            old_var_name = var_name
            var_name = var_name.replace(src_pattern, tgt_pattern)
            logger.info("Converted: %s --> %s", old_var_name, var_name)
    return var_name


def _has_exclude_patterns(name, exclude_patterns):
    """Checks if a string contains substrings that match patterns to exclude."""
    for p in exclude_patterns:
        if p in name:
            return True
    return False


def _get_permutation(name, permutations):
    """Checks whether a variable requires transposition by pattern matching."""
    for src_pattern, permutation in permutations:
        if src_pattern in name:
            logger.info("Permuted: %s --> %s", name, permutation)
            return permutation
    return None


def _get_new_shape(name, shape, num_heads):
    """Checks whether a variable requires reshape by pattern matching."""
    if "self_attention/attention_output/kernel" in name:
        return (num_heads, shape[0] // num_heads, shape[1])
    if "self_attention_output/bias" in name:
        return shape

    patterns = ["self_attention/query", "self_attention/value", "self_attention/key"]
    for pattern in patterns:
        if pattern in name:
            if "kernel" in name:
                return tuple([shape[0], num_heads, shape[1] // num_heads])
            if "bias" in name:
                return tuple([num_heads, shape[0] // num_heads])
    return None


def convert_checkpoint(
    checkpoint_from_path: str,
    checkpoint_to_path: str,
    num_heads: int,
    original_root_name: str = "bert",
    converted_root_name: str = "bert",
    name_replacements: Optional[Sequence] = None,
    permutations: Optional[Sequence] = None,
    exclude_patterns: Optional[Sequence] = None,
) -> None:
    """Migrates the names of variables within a checkpoint.
    Args:
      checkpoint_from_path: Path to source checkpoint to be read in.
      checkpoint_to_path: Path to checkpoint to be written out.
      num_heads: The number of heads of the model.
      name_replacements: A list of tuples of the form (match_str, replace_str)
        describing variable names to adjust.
      permutations: A list of tuples of the form (match_str, permutation)
        describing permutations to apply to given variables. Note that match_str
        should match the original variable name, not the replaced one.
      exclude_patterns: A list of string patterns to exclude variables from
        checkpoint conversion.
    Returns:
      A dictionary that maps the new variable names to the Variable objects.
      A dictionary that maps the old variable names to the new variable names.
    """
    name_replacements = list(name_replacements) or list()
    if converted_root_name != "":
        name_replacements.insert(0, (original_root_name, converted_root_name))
    else:
        name_replacements.insert(0, (original_root_name + "/", ""))
    permutations = permutations or list()
    exclude_patterns = exclude_patterns or ["adam", "Adam"]
    with tfv1.Graph().as_default():
        logger.info("Reading checkpoint_from_path %s", checkpoint_from_path)
        reader = tfv1.train.load_checkpoint(checkpoint_from_path)
        name_shape_map = reader.get_variable_to_shape_map()
        new_variable_map = {}
        conversion_map = {}
        for var_name in name_shape_map:
            if exclude_patterns and _has_exclude_patterns(var_name, exclude_patterns):
                continue
            # Get the original tensor data.
            tensor = reader.get_tensor(var_name)

            # Look up the new variable name, if any.
            new_var_name = _bert_name_replacement(var_name, name_replacements)

            # See if we need to reshape the underlying tensor.
            new_shape = None
            if num_heads > 0:
                new_shape = _get_new_shape(new_var_name, tensor.shape, num_heads)
            if new_shape:
                logger.info(
                    "Veriable %s has a shape change from %s to %s",
                    var_name,
                    tensor.shape,
                    new_shape,
                )
                tensor = np.reshape(tensor, new_shape)

            # See if we need to permute the underlying tensor.
            permutation = _get_permutation(var_name, permutations)
            if permutation:
                tensor = np.transpose(tensor, permutation)

            # Create a new variable with the possibly-reshaped or transposed tensor.
            var = tfv1.Variable(tensor, name=var_name)

            # Save the variable into the new variable map.
            new_variable_map[new_var_name] = var

            # Keep a list of converter variables for sanity checking.
            if new_var_name != var_name:
                conversion_map[var_name] = new_var_name

        saver = tfv1.train.Saver(new_variable_map)

        with tfv1.Session() as sess:
            sess.run(tfv1.global_variables_initializer())
            logger.info("Writing checkpoint_to_path %s", checkpoint_to_path)
            saver.save(sess, checkpoint_to_path, write_meta_graph=True)

    logger.info("Summary:")
    logger.info("  Converted %d variable name(s).", len(new_variable_map))
    logger.info("  Converted: %s", str(conversion_map))


convert_bert_checkpoint = partial(
    convert_checkpoint,
    name_replacements=BERT_NAME_REPLACEMENTS,
    permutations=BERT_PERMUTATIONS,
)
convert_electra_checkpoint = partial(
    convert_bert_checkpoint, original_root_name="electra"
)
